from django.contrib.auth import get_user_model
from django.core.mail import EmailMultiAlternatives
from django.utils import timezone
from django_slack import slack_message
from firebase_admin.firestore import client
from firebase_admin.messaging import Message
from firebase_admin.messaging import Notification
from otp_twilio.models import TwilioSMSDevice
from typing import Dict
from typing import List

import json
import logging
import requests


User = get_user_model()


class BaseChannel:
    template_parts = ["subject.txt", "body.txt", "body.html", "data.json"]

    def send(self, users: List[User], context: Dict[str, str]):
        pass


class EmailChannel(BaseChannel):
    template_parts = ["subject.txt", "body.txt", "body.html"]

    def send(self, users: List[User], context: Dict[str, str]):
        recipients = context.get(
            "recipients", [user.email for user in users if user.email]
        )
        msg = EmailMultiAlternatives(
            subject=context["subject.txt"], to=recipients, body=context["body.txt"]
        )
        msg.attach_alternative(context["body.html"], "text/html")
        for attachment in context.get("attachments", []):
            msg.attach(**attachment)
        msg.send()


class ConsoleChannel(BaseChannel):
    template_parts = ["subject.txt", "body.txt"]

    def send(self, users: List[User], context: Dict[str, str]):
        logging.info(
            f"users: {users}; subject: {context['subject.txt']}; body: {context['body.txt']}"
        )


class FirebasePushChannel(BaseChannel):
    template_parts = ["subject.txt", "body.txt", "data.json"]

    def send(self, users: List[User], context: Dict[str, str]):
        try:
            devices = context["devices_queryset"]
        except KeyError:
            from df_notifications.models import UserDevice

            devices = UserDevice.objects.all()

        devices.filter(user__in=users).send_message(
            Message(
                notification=Notification(
                    title=context["subject.txt"],
                    body=context["body.txt"],
                ),
                data=json.loads(context["data.json"]),
            ),
        )


class JSONPostWebhookChannel(BaseChannel):
    template_parts = ["data.json"]

    def send(self, users: List[User], context: Dict[str, str]):
        requests.post(context["url"], json=json.loads(context["data.json"]))


class SlackChannel(BaseChannel):
    template_parts = ["subject.txt", "body.txt"]

    def send(self, users: List[User], context: Dict[str, str]):
        slack_message(
            "df_notifications/base_slack.html",
            context=context,
            attachments=[
                {"title": context["subject.txt"], "text": context["body.txt"]}
            ],
        )


class FirebaseChatChannel(BaseChannel):
    template_parts = ["body.txt"]

    def send(self, users: List[User], context: Dict[str, str]):
        db = client()
        db.collection("rooms").document(context["chat_room_id"]).collection(
            "messages"
        ).document().set(
            {
                "text": context["body.txt"],
                "createdAt": timezone.now(),
                "updatedAt": timezone.now(),
                "type": "text",
                "authorId": context.get("chat_author_id", "system"),
            }
        )


class TwilioSMSChannel(BaseChannel):
    template_parts = ["body.txt"]

    def send(self, users: List[User], context: Dict[str, str]):
        for device in TwilioSMSDevice.objects.filter(user__in=users):
            device._deliver_token(context["body.txt"])



# Generated by Django 4.1.5 on 2023-01-26 10:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0004_alter_postnotificationrule_is_published_prev'),
    ]

    operations = [
        migrations.RenameField(
            model_name='postnotificationreminder',
            old_name='template_prefix',
            new_name='template',
        ),
        migrations.RenameField(
            model_name='postnotificationrule',
            old_name='template_prefix',
            new_name='template',
        ),
        migrations.AlterField(
            model_name='postnotificationreminder',
            name='action',
            field=models.TextField(blank=True, default='', help_text='Python code to execute. You can use `instance` variable to access current model'),
        ),
    ]



# Generated by Django 4.1.5 on 2023-01-17 11:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0002_alter_postnotificationreminder_history_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='postnotificationreminder',
            name='action',
            field=models.TextField(default='', help_text='Python code to execute. You can use `instance` variable to access current model'),
        ),
    ]



from dbtemplates.models import Template
from df_notifications.models import NotificationHistory
from df_notifications.tasks import send_notification_async
from df_notifications.utils import send_notification
from django.contrib.auth import get_user_model
from django.utils import timezone
from tests.test_app.models import Post
from tests.test_app.models import PostNotificationReminder
from tests.test_app.models import PostNotificationRule

import pytest


User = get_user_model()

pytestmark = [pytest.mark.django_db]


def setup_templates():
    Template.objects.create(
        name="df_notifications/posts/published/subject.txt",
        content="New post: {{ instance.title }}",
    )
    Template.objects.create(
        name="df_notifications/posts/published/body.txt",
        content="{{ instance.description }}",
    )


def setup_published_notification():
    action = PostNotificationRule(
        is_published_next=True,
        is_published_prev=False,
        channel="console",
        template_prefix="df_notifications/posts/published/",
    )
    action.save()


def test_post_published_notification_created():
    setup_published_notification()
    setup_templates()

    user = User.objects.create(
        email="test@test.com",
    )
    post = Post.objects.create(
        title="Title 1",
        description="Content 1",
        is_published=True,
        author=user,
    )
    notifications = NotificationHistory.objects.all()
    assert len(notifications) == 1
    notification = notifications[0]
    assert notification.content["subject.txt"] == f"New post: {post.title}"
    assert notification.content["body.txt"] == post.description


def test_post_non_published_notification_not_created():
    setup_published_notification()
    setup_templates()

    user = User.objects.create(
        email="test@test.com",
    )
    Post.objects.create(
        title="Title 1",
        description="Content 1",
        is_published=False,
        author=user,
    )
    assert not NotificationHistory.objects.exists()


def test_notification_history_queryset_for_instance_filtering():
    setup_published_notification()
    setup_templates()

    user = User.objects.create(
        email="test@test.com",
    )
    post1 = Post.objects.create(
        title="Title 1",
        description="Content 1",
        is_published=True,
        author=user,
    )
    Post.objects.create(
        title="Title 2",
        description="Content 2",
        is_published=True,
        author=user,
    )

    notifications = NotificationHistory.objects.for_instance(post1)
    assert len(notifications) == 1
    notification = notifications[0]
    assert notification.content["body.txt"] == post1.description


def test_reminder_queryset_is_valid():
    setup_templates()
    reminder = PostNotificationReminder(
        is_published=True,
        channel="console",
        template_prefix="df_notifications/posts/published/",
    )
    reminder.save()

    user = User.objects.create(
        email="test@test.com",
    )
    post = Post.objects.create(
        title="Title 1",
        description="Content 1",
        is_published=True,
        author=user,
    )

    posts = list(reminder.get_model_queryset())
    assert len(posts) == 1
    assert posts[0] == post


def test_reminder_respect_delay():
    setup_templates()
    reminder = PostNotificationReminder(
        is_published=True,
        channel="console",
        template_prefix="df_notifications/posts/published/",
        delay=timezone.timedelta(seconds=60),
    )
    reminder.save()

    user = User.objects.create(
        email="test@test.com",
    )
    Post.objects.create(
        title="Title 1",
        description="Content 1",
        is_published=True,
        author=user,
    )
    assert not reminder.get_model_queryset().exists()

    Post.objects.update(updated=timezone.now() - timezone.timedelta(seconds=120))
    assert reminder.get_model_queryset().exists()


def test_reminder_performs_retries():
    setup_templates()
    reminder = PostNotificationReminder(
        is_published=True,
        channel="console",
        template_prefix="df_notifications/posts/published/",
        repeat=2,
        cooldown=timezone.timedelta(seconds=0),
        delay=timezone.timedelta(seconds=0),
    )
    reminder.save()

    user = User.objects.create(
        email="test@test.com",
    )
    post = Post.objects.create(
        title="Title 1",
        description="Content 1",
        is_published=True,
        author=user,
    )

    PostNotificationReminder.invoke()
    assert post.notifications.count() == 1
    PostNotificationReminder.invoke()
    assert post.notifications.count() == 2
    PostNotificationReminder.invoke()
    assert post.notifications.count() == 2


def test_send_notification_without_model():
    Template.objects.create(
        name="df_notifications/posts/published/subject.txt",
        content="New post: {{ title }}",
    )
    Template.objects.create(
        name="df_notifications/posts/published/body.txt",
        content="{{ description }}",
    )
    user = User.objects.create(
        email="test@test.com",
    )

    notification = send_notification(
        users=[user],
        channel="console",
        template_prefixes="df_notifications/posts/published/",
        context={
            "title": "title 123",
            "description": "description 456",
        },
    )
    assert notification.content["subject.txt"] == "New post: title 123"
    assert notification.content["body.txt"] == "description 456"


def test_send_notification_async_without_model():
    user = User.objects.create(
        email="test@test.com",
    )
    Template.objects.create(
        name="df_notifications/posts/published/subject.txt",
        content="New post: {{ title }}",
    )
    Template.objects.create(
        name="df_notifications/posts/published/body.txt",
        content="{{ description }}",
    )

    send_notification_async(
        [user.id],
        "console",
        "df_notifications/posts/published/",
        {
            "title": "title 123",
            "description": "description 456",
        },
    )
    notifications = NotificationHistory.objects.all()
    assert len(notifications) == 1
    notification = notifications[0]
    assert notification.content["subject.txt"] == "New post: title 123"
    assert notification.content["body.txt"] == "description 456"


def test_reminder_performs_action():
    setup_templates()
    reminder = PostNotificationReminder(
        is_published=True,
        channel="console",
        template_prefix="df_notifications/posts/published/",
        action="instance.title='new title'; instance.save()",
    )
    reminder.save()

    user = User.objects.create(
        email="test@test.com",
    )
    post = Post.objects.create(
        title="Title 1",
        description="Content 1",
        is_published=True,
        author=user,
    )
    PostNotificationReminder.invoke()
    post.refresh_from_db()
    assert post.title == "new title"


def test_rule_not_invoked_on_non_tracked_field_change():
    setup_templates()
    action = PostNotificationRule(
        is_published_next=True,
        is_published_prev=None,
        channel="console",
        template_prefix="df_notifications/posts/published/",
    )
    action.save()

    user = User.objects.create(
        email="test@test.com",
    )
    post = Post.objects.create(
        title="Title 1",
        description="Content 1",
        is_published=True,
        author=user,
    )
    assert len(NotificationHistory.objects.all()) == 1
    post.title = "Title 2"
    post.save()
    assert len(NotificationHistory.objects.all()) == 1
    post.is_published = False
    post.save()
    assert len(NotificationHistory.objects.all()) == 1
    post.is_published = True
    post.save()
    assert len(NotificationHistory.objects.all()) == 2


def test_default_templates_rendered_if_no_template_exists():
    Template.objects.create(
        name="test_app/df_notifications/posts/subject.txt",
        content="New post: {{ instance.title }}",
    )
    Template.objects.create(
        name="test_app/df_notifications/posts/body.txt",
        content="{{ instance.description }}",
    )
    setup_published_notification()

    user = User.objects.create(
        email="test@test.com",
    )
    post = Post.objects.create(
        title="Title 1",
        description="Content 1",
        is_published=True,
        author=user,
    )
    notifications = NotificationHistory.objects.all()
    assert len(notifications) == 1
    notification = notifications[0]
    assert notification.content["subject.txt"] == f"New post: {post.title}"
    assert notification.content["body.txt"] == post.description



from datetime import timedelta
from df_notifications.fields import NoMigrationsChoicesField
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db import transaction
from django.db.models import Count
from django.db.models import Q
from django.db.models import QuerySet
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from fcm_django.models import AbstractFCMDevice
from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar


M = TypeVar("M", bound=models.Model)

User = get_user_model()

# https://code.djangoproject.com/ticket/33174
if TYPE_CHECKING:

    class GenericBase(Generic[M]):
        pass

else:

    class GenericBase:
        def __class_getitem__(cls, _):
            return cls


class UserDevice(AbstractFCMDevice):
    user = models.ForeignKey(
        User,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_query_name=_("fcmdevice"),
    )

    class Meta:
        verbose_name = _("User device")
        verbose_name_plural = _("User devices")


# -------- Notifications ----------


class NotificationHistoryQuerySet(models.QuerySet):
    def for_instance(self, instance: M):
        return self.filter(
            instance_id=instance.pk,
            content_type=ContentType.objects.get_for_model(instance).id,
        )


class NotificationHistory(models.Model):
    id = models.BigAutoField(primary_key=True)
    users = models.ManyToManyField(User, blank=True)
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    channel = NoMigrationsChoicesField(
        max_length=255,
        choices=[(key, key) for key in settings.DF_NOTIFICATIONS["CHANNELS"]],
    )
    template_prefix = models.CharField(max_length=255)
    content = models.JSONField(default=dict, blank=True)

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, null=True, blank=True
    )
    instance_id = models.PositiveBigIntegerField(null=True, blank=True)
    instance = GenericForeignKey("content_type", "instance_id")

    objects = NotificationHistoryQuerySet.as_manager()

    class Meta:
        verbose_name_plural = "Notification history"

        indexes = [
            models.Index(fields=["template_prefix", "created"]),
            models.Index(fields=["channel", "created"]),
            models.Index(fields=["content_type", "instance_id", "created"]),
        ]


class NotifiableModelMixin(models.Model):
    notifications = GenericRelation(NotificationHistory, "instance_id")

    class Meta:
        abstract = True


# ----------- Actions -------------


class BaseModelRule(GenericBase[M], models.Model):
    model: Type[M]
    tracking_fields: Optional[List[str]] = None

    @classmethod
    def is_field_changed(cls, instance: M, prev: Optional[M]):
        if cls.tracking_fields is None or prev is None:
            return True

        for field in cls.tracking_fields:
            if getattr(instance, field) != getattr(prev, field):
                return True

        return False

    @classmethod
    def get_queryset(cls, instance: M, prev: Optional[M]) -> QuerySet["BaseModelRule"]:
        return cls.objects.all()

    def check_condition(self, instance: M, prev: Optional[M]) -> bool:
        return True

    def perform_action(self, instance: M):
        pass

    @classmethod
    def invoke(cls, instance: M):
        prev = getattr(instance, "_pre_save_instance", None)

        if not cls.is_field_changed(instance, prev):
            return

        for action in cls.get_queryset(instance, prev):
            if action.check_condition(instance, prev):
                action.perform_action(instance)

    class Meta:
        abstract = True


class BaseModelReminder(GenericBase[M], models.Model):
    model: Type[M]

    @classmethod
    def get_queryset(cls) -> QuerySet["BaseModelReminder"]:
        return cls.objects.all()

    def get_model_queryset(self) -> QuerySet[M]:
        return self.model.objects.all()

    def check_condition(self, instance: M) -> bool:
        return True

    def perform_action(self, instance: M):
        pass

    @classmethod
    def invoke(cls):
        for reminder in cls.get_queryset():
            for instance in reminder.get_model_queryset():
                if reminder.check_condition(instance):
                    reminder.perform_action(instance)

    class Meta:
        abstract = True


class NotificationModelMixin(models.Model):
    model: Type[M]
    admin_list_display: List[str] = ["channel", "template_prefix"]

    history = models.ManyToManyField(NotificationHistory, blank=True, editable=False)
    channel = NoMigrationsChoicesField(
        max_length=255,
        choices=[(key, key) for key in settings.DF_NOTIFICATIONS["CHANNELS"]],
    )
    template_prefix = models.CharField(max_length=255)
    context = models.JSONField(default=dict, blank=True)

    def get_users(self, instance: M) -> List[User]:
        return []

    def get_context(self, instance: M) -> Dict[str, Any]:
        return {
            **self.context,
            "instance": instance,
        }

    def get_template_prefixes(self):
        return [
            self.template_prefix,
            f"{self.model._meta.app_label}/df_notifications/{self.model._meta.model_name}",
        ]

    def send(self, instance: M):
        from .utils import send_notification

        notification = send_notification(
            self.get_users(instance),
            self.channel,
            self.get_template_prefixes(),
            self.get_context(instance),
        )
        self.history.add(notification)

    class Meta:
        abstract = True


class AsyncNotificationModelMixin(NotificationModelMixin):
    def send(self, instance: M) -> User:
        from .tasks import send_model_notification_async

        transaction.on_commit(
            lambda: send_model_notification_async.delay(
                self._meta.label_lower,
                str(self.pk),
                str(instance.pk),
            )
        )

    class Meta:
        abstract = True


class NotificationModelRule(NotificationModelMixin, BaseModelRule):
    def perform_action(self, instance: M):
        self.send(instance)

    class Meta:
        abstract = True


class NotificationModelReminder(NotificationModelMixin, BaseModelReminder):
    MODIFIED_MODEL_FIELD = "modified"

    delay = models.DurationField(
        help_text="Send the reminder after this period of time",
        default=timedelta(seconds=0),
    )
    cooldown = models.DurationField(
        help_text="Wait so much time before reminding again",
        default=timedelta(hours=1),
    )
    repeat = models.SmallIntegerField(
        help_text="Repeat the reminder this many times", default=1
    )
    action = models.TextField(
        help_text="Python code to execute. You can use "
        "`instance` variable to access current model",
        default="",
        blank=True,
    )

    def get_model_queryset(self) -> QuerySet[M]:
        return (
            super(NotificationModelReminder, self)
            .get_model_queryset()
            .filter(**{f"{self.MODIFIED_MODEL_FIELD}__lt": timezone.now() - self.delay})
            .annotate(
                notification_count=Count(
                    "notifications__id",
                    filter=Q(notifications__in=self.history.all()),
                ),
            )
            .filter(Q(notification_count__lt=self.repeat))
            .exclude(
                Q(notifications__in=self.history.all())
                & Q(notifications__created__gt=(timezone.now() - self.cooldown))
            )
            .distinct()
        )

    def perform_action(self, instance: M):
        self.send(instance)
        if self.action:
            exec(self.action)

    class Meta:
        abstract = True


class NotificationModelAsyncRule(AsyncNotificationModelMixin, NotificationModelRule):
    class Meta:
        abstract = True


class NotificationModelAsyncReminder(
    AsyncNotificationModelMixin, NotificationModelReminder
):
    class Meta:
        abstract = True



from .models import NotificationHistory
from .models import UserDevice
from django.contrib import admin
from fcm_django.admin import DeviceAdmin
from fcm_django.models import FCMDevice


admin.site.unregister(FCMDevice)


@admin.register(UserDevice)
class UserDeviceAdmin(DeviceAdmin):
    pass


@admin.register(NotificationHistory)
class NotificationHistoryAdmin(admin.ModelAdmin):
    list_display = ("template_prefix", "channel", "instance_id", "created")
    date_hierarchy = "created"
    search_fields = (
        "template_prefix",
        "channel",
    )



# Generated by Django 4.1.5 on 2023-01-18 05:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0003_postnotificationreminder_action'),
    ]

    operations = [
        migrations.AlterField(
            model_name='postnotificationrule',
            name='is_published_prev',
            field=models.BooleanField(default=False, null=True),
        ),
    ]



# Generated by Django 4.1.5 on 2023-01-16 09:31

import datetime
import df_notifications.fields
import df_notifications.models
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('df_notifications', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='PostNotificationRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('channel', df_notifications.fields.NoMigrationsChoicesField(max_length=255)),
                ('template_prefix', models.CharField(max_length=255)),
                ('context', models.JSONField(blank=True, default=dict)),
                ('is_published_prev', models.BooleanField(default=False)),
                ('is_published_next', models.BooleanField(default=True)),
                ('history', models.ManyToManyField(blank=True, to='df_notifications.notificationhistory')),
            ],
            options={
                'abstract': False,
            },
            bases=(df_notifications.models.GenericBase, models.Model),
        ),
        migrations.CreateModel(
            name='PostNotificationReminder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('channel', df_notifications.fields.NoMigrationsChoicesField(max_length=255)),
                ('template_prefix', models.CharField(max_length=255)),
                ('context', models.JSONField(blank=True, default=dict)),
                ('delay', models.DurationField(default=datetime.timedelta(0), help_text='Send the reminder after this period of time')),
                ('cooldown', models.DurationField(default=datetime.timedelta(seconds=3600), help_text='Wait so much time before reminding again')),
                ('repeat', models.SmallIntegerField(default=1, help_text='Repeat the reminder this many times')),
                ('is_published', models.BooleanField(default=True)),
                ('history', models.ManyToManyField(blank=True, to='df_notifications.notificationhistory')),
            ],
            options={
                'abstract': False,
            },
            bases=(df_notifications.models.GenericBase, models.Model),
        ),
        migrations.CreateModel(
            name='Post',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255)),
                ('description', models.TextField()),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('is_published', models.BooleanField(default=False)),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# Generated by Django 4.1.6 on 2023-02-22 09:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0005_rename_template_prefix_postnotificationreminder_template_and_more'),
    ]

    operations = [
        migrations.RenameField(
            model_name='postnotificationreminder',
            old_name='template',
            new_name='template_prefix',
        ),
        migrations.RenameField(
            model_name='postnotificationrule',
            old_name='template',
            new_name='template_prefix',
        ),
    ]



from df_notifications.models import NotificationModelMixin
from df_notifications.utils import get_channel_instance
from django.contrib import admin
from django.db import transaction
from django.db.models.signals import post_save
from django.db.models.signals import pre_save
from django.template import TemplateDoesNotExist
from django.template.loader import get_template
from django.urls import reverse
from django.utils.safestring import mark_safe
from import_export.admin import ImportExportMixin
from import_export.resources import ModelResource


def save_previous_instance(sender, instance, **kwargs):
    if instance.pk:
        instance._pre_save_instance = sender.objects.get(pk=instance.pk)
    else:
        instance._pre_save_instance = None


def create_proxy_model(model_class):
    return type(
        model_class.__name__,
        (model_class,),
        {
            "__module__": "df_notifications",
            "Meta": type("Meta", (object,), {"proxy": True, "auto_created": True}),
        },
    )


def register_notification_model_admin(model_class):
    ProxyModel = create_proxy_model(model_class)

    @admin.register(ProxyModel)
    class AdminProxyModel(ImportExportMixin, admin.ModelAdmin):
        def get_list_display(self, request):
            return (
                *model_class.admin_list_display,
                "content",
            )

        def get_resource_classes(self):
            class ProxyModelResource(ModelResource):
                class Meta:
                    model = model_class
                    skip_unchanged = True
                    report_skipped = False

            return [ProxyModelResource]

        def content(self, obj):
            url = reverse("admin:dbtemplates_template_changelist")
            return mark_safe(f'<a href="{url}?q={obj.template_prefix}">Change</a>')

        @admin.action(
            description="Initialize DB Templates for notifications (override current content)"
        )
        def populate(self, request, qs):
            from dbtemplates.models import Template

            for item in qs:
                assert isinstance(item, NotificationModelMixin)

                for part in get_channel_instance(item.channel).template_parts:
                    names = [
                        f"{item.template_prefix}{item.channel}__{part}",
                        f"{item.template_prefix}{part}",
                    ]
                    for name in names:
                        try:
                            with transaction.atomic():
                                Template.objects.filter(name=name).delete()
                                template = get_template(name)
                                with open(template.origin.name, "r") as f:
                                    Template.objects.create(name=name, content=f.read())
                        except TemplateDoesNotExist:
                            pass

        actions = [populate]


def register_rule_model(rule_class):
    pre_save.connect(
        save_previous_instance,
        rule_class.model,
        weak=False,
        dispatch_uid="save_previous_instance",
    )

    def apply_action(sender, instance, **kwargs):
        rule_class.invoke(instance)

    post_save.connect(apply_action, rule_class.model, weak=False)

    register_notification_model_admin(rule_class)

    return rule_class


def register_reminder_model(reminder_class):
    register_notification_model_admin(reminder_class)
    return reminder_class
